// TI-HIP (Hide in Picture) Version 2.3
// Copyright  2001, Salmonsoft TI Solutions
// Salmon1231@excite.com

#define OPTIMIZE_ROM_CALLS
#define SAVE_SCREEN
#include <tigcclib.h>

#define swap(a, b) a ^= b ^= a ^= b

short int i, j, _ti92plus;
unsigned char S[256];

unsigned char cryptByte(const unsigned char c);
void cryptSetup(const unsigned char *key, const short int keylen);
void fileIOError(void);
void fileSizeError(void);
void getPassword(void);
void hide(const char *filename);
unsigned char highByte(const short int in);
unsigned char lowByte(const short int in);
void mainDialog(char *filename, short int *mode);
short int makeShort(const unsigned char highbyte, const unsigned char lowbyte);
void memError(void);
FILE *openPic(const char *filename);
FILE *openVar(const char *filename, const unsigned char mode);
void seek(const char *filename);

unsigned char cryptByte(const unsigned char c)
{
	i = (i + 1) % 256;
	j = (j + S[i]) % 256;
	swap(S[i], S[j]);
	return c ^ S[(S[i] + S[j]) % 256];
}

void cryptSetup(const unsigned char key[], const short int keylen)
{
	unsigned char K[256];
	for(i = 0, j = 0; i < 256; i++)
	{
		S[i] = i;
		K[i] = key[j];
		j = (j + 1 == keylen ? 0 : j + 1);
	}
	for(i = 0, j = 0; i < 256; i++)
	{
		j = (j + S[i] + K[i]) % 256;
		swap(S[i], S[j]);
	}
	i = 0;
	j = 0;
}

void fileIOError(void)
{
	DlgMessage("ERROR", "File Error or File not Found", BT_OK, BT_NONE);
	exit(0);
}

void fileSizeError(void)
{
	DlgMessage("ERROR", "File too large to be hidden", BT_OK, BT_NONE);
	exit(0);
}

void getPassword(void)
{
	LCD_BUFFER screen;
	short int character;
	unsigned char i, password[31], stars[31];
	unsigned long int hashval;
	for(i = 0; i < 31; i++)
	{
		password[i] = 0;
		stars[i] = 0;
	}
	LCD_save(screen);
	clrscr();
	for(i = 0; i < 30; i++)
	{
		MoveTo(0, 0);
		printf("Password: %s  ", stars);
		character = ngetchx();
		if(character == KEY_ENTER)
			break;
		if(character == KEY_BACKSPACE)
		{
			if(i > 0)
			{
				i--;
				password[i] = 0;
				stars[i] = 0;
			}
			i--;
			continue;
		}
		password[i] = character;
		stars[i] = '*';
	}
	LCD_restore(screen);
	cryptSetup(password, strlen(password));
	hashval = 0;
	for(i = 0; i < strlen(password); i++)
		hashval = 317 * hashval + password[i];
	srand(hashval);
}

void hide(const char *filename)
{
	FILE *pic, *var;
	short int i, offset, picsize, varsize;
	unsigned char *usedspace;
	pic = openPic(filename);
	if(pic == NULL)
		fileIOError();
	for(picsize = 0; getc(pic) != EOF; picsize++);
	fclose(pic);
	picsize -= 5;
	var = openVar(filename, TRUE);
	if(var == NULL)
		fileIOError();
	for(varsize = 0; getc(var) != EOF; varsize++);
	fclose(var);
	if(picsize - 2 < varsize)
		fileSizeError();
	usedspace = (unsigned char *)malloc(picsize);
	if(usedspace == NULL)
		memError();
	for(i = 0; i < picsize; i++)
		usedspace[i] = 0;
	getPassword();
	pic = openPic(filename);
	var = openVar(filename, TRUE);
	while(usedspace[offset = random(picsize)]);
	fseek(pic, offset + 4, SEEK_SET);
	putc(highByte(varsize), pic);
	usedspace[offset] = 1;
	while(usedspace[offset = random(picsize)]);
	fseek(pic, offset + 4, SEEK_SET);
	putc(lowByte(varsize), pic);
	usedspace[offset] = 1;
	for(i = 0; i < varsize; i++)
	{
		while(usedspace[offset = random(picsize)]);
		fseek(pic, offset + 4, SEEK_SET);
		putc(cryptByte(getc(var)), pic);
		usedspace[offset] = 1;
	}
	fclose(pic);
	fclose(var);
	free(usedspace);
}

unsigned char highByte(const short int in)
{
	return (in - in % 256) / 256;
}

unsigned char lowByte(const short int in)
{
	return in % 256;
}

void mainDialog(char *filename, short int *mode)
{
	HANDLE dialog, menu;
	short int status;
	if((menu = PopupNew(NULL, 0)) == H_NULL)
		exit(0);
	if(PopupAddText(menu, -1, "Hide to PIC", 1) == H_NULL)
	{
		HeapFree(menu);
		exit(0);
	}
	if(PopupAddText(menu, -1, "Retrieve from PIC", 2) == H_NULL)
	{
		HeapFree(menu);
		exit(0);
	}
	if((dialog = DialogNewSimple(230, 110)) == H_NULL)
	{
		HeapFree(menu);
		exit(0);
	}
	if(DialogAddTitle(dialog, "TI-HIP (Hide in Picture) Version 2.3", BT_OK, BT_CANCEL) == H_NULL)
	{
		HeapFree(menu);
		HeapFree(dialog);
		exit(0);
	}
	if(DialogAddPulldown(dialog, 10, 20, "Mode   :", menu, 0) == H_NULL)
	{
		HeapFree(menu);
		HeapFree(dialog);
		exit(0);
	}
	if(DialogAddRequest(dialog, 10, 32, "File   :", 0, 17, 16) == H_NULL)
	{
		HeapFree(menu);
		HeapFree(dialog);
		exit(0);
	}
	if(DialogAddRequest(dialog, 10, 44, "Picture:", 18, 17, 16) == H_NULL)
	{
		HeapFree(menu);
		HeapFree(dialog);
		exit(0);
	}
	if(DialogAddText(dialog, 10, 66, " 2001, Salmonsoft TI Solutions") == H_NULL)
	{
		HeapFree(menu);
		HeapFree(dialog);
		exit(0);
	}
	if(DialogAddText(dialog, 10, 78, "Salmon1231@excite.com") == H_NULL)
	{
		HeapFree(menu);
		HeapFree(dialog);
		exit(0);
	}
	status = DialogDo(dialog, CENTER, CENTER, filename, mode);
	HeapFree(menu);
	HeapFree(dialog);
	if(status == KEY_ESC)
		exit(0);
}

short int makeShort(const unsigned char highbyte, const unsigned char lowbyte)
{
	return highbyte * 256 + lowbyte;
}

void memError(void)
{
	DlgMessage("ERROR", "Out of Memory", BT_OK, BT_NONE);
	exit(0);
}

FILE *openPic(const char *filename)
{
	FILE *pic;
	short int cols, rows;
	pic = fopen(filename + 18, "r+b");
	if(pic == NULL)
		return NULL;
	if(getc(pic) != 0)
		return NULL;
	if((rows = getc(pic)) > 103)
		return NULL;
	if(getc(pic) != 0)
		return NULL;
	if((cols = getc(pic)) > 239)
		return NULL;
	fseek(pic, -1, SEEK_END);
	if(getc(pic) != 223)
		return NULL;
	fseek(pic, 0, SEEK_SET);
	return pic;
}

FILE *openVar(const char *filename, const unsigned char mode)
{
	FILE *var;
	if(mode)
		var = fopen(filename, "rb");
	else
		var = fopen(filename, "wb");
	if(var == NULL)
		return NULL;
	return var;
}

void seek(const char *filename)
{
	FILE *pic, *var;
	short int i, offset, picsize, varsize;
	unsigned char highbyte, lowbyte, *usedspace;
	pic = openPic(filename);
	if(pic == NULL)
		fileIOError();
	for(picsize = 0; getc(pic) != EOF; picsize++);
	fclose(pic);
	picsize -= 5;
	usedspace = (unsigned char *)malloc(picsize);
	if(usedspace == NULL)
		memError();
	for(i = 0; i < picsize; i++)
		usedspace[i] = 0;
	getPassword();
	pic = openPic(filename);
	while(usedspace[offset = random(picsize)]);
	fseek(pic, offset + 4, SEEK_SET);
	highbyte = getc(pic);
	usedspace[offset] = 1;
	while(usedspace[offset = random(picsize)]);
	fseek(pic, offset + 4, SEEK_SET);
	lowbyte = getc(pic);
	usedspace[offset] = 1;
	varsize = makeShort(highbyte, lowbyte);
	if(picsize - 2 < varsize)
	{
		fclose(pic);
		fileIOError();
	}
	var = openVar(filename, FALSE);
	if(var == NULL)
	{
		fclose(pic);
		fileIOError();
	}
	for(i = 0; i < varsize; i++)
	{
		while(usedspace[offset = random(picsize)]);
		fseek(pic, offset + 4, SEEK_SET);
		putc(cryptByte(getc(pic)), var);
		usedspace[offset] = 1;
	}
	fclose(pic);
	fclose(var);
	free(usedspace);
}

void _main(void)
{
	char filename[36];
	short int i, mode[1];
	for(i = 0; i < 36; i++)
		filename[i] = 0;
	mode[0] = 1;
	mainDialog(filename, mode);
	if(mode[0] == 1)
		hide(filename);
	if(mode[0] == 2)
		seek(filename);
}
